""" test_slot.py - test asyncSlot with signal-slot connection """

import asyncio
import unittest
from shim import QtCore
from asyncslot import asyncSlot, AsyncSlotRunner


called = []


def visit(s, tag=None):
    if tag is not None:
        msg = f'{s}({tag.secret})'
    else:
        msg = s
    # print(msg)
    called.append(msg)


class MySignalObject(QtCore.QObject):
    ready0 = QtCore.Signal()
    ready1 = QtCore.Signal(bool)


class MySlotMixin:
    pass


qt_slot_supports_descriptor = not QtCore.__name__.startswith('PyQt')


class MySlotObject(QtCore.QObject):
    secret = 'Cls'

    def __init__(self):
        super().__init__()
        self.secret = 'Self'

    # -------------------------------------------------------------------------
    # Instance method
    # -------------------------------------------------------------------------

    def method(self):
        visit('method', self)

    @QtCore.Slot()
    def slot_method(self):
        visit('slot_method', self)

    async def amethod(self):
        visit('amethod.1', self)
        await asyncio.sleep(0)
        visit('amethod.2', self)

    @QtCore.Slot()
    async def slot_amethod(self):
        visit('slot_amethod.1', self)
        await asyncio.sleep(0)
        visit('slot_amethod.2', self)

    @asyncSlot
    async def decorated_amethod(self):
        visit('decorated_amethod.1', self)
        await asyncio.sleep(0)
        visit('decorated_amethod.2', self)

    @QtCore.Slot()
    @asyncSlot
    async def slot_decorated_amethod(self):
        visit('slot_decorated_amethod.1', self)
        await asyncio.sleep(0)
        visit('slot_decorated_amethod.2', self)

    @asyncSlot
    @QtCore.Slot()
    async def decorated_slot_amethod(self):
        visit('decorated_slot_amethod.1', self)
        await asyncio.sleep(0)
        visit('decorated_slot_amethod.2', self)

    # -------------------------------------------------------------------------
    # Class method
    # -------------------------------------------------------------------------

    @classmethod
    def class_method(cls):
        visit('class_method', cls)

    @classmethod
    async def class_amethod(cls):
        visit('class_amethod.1', cls)
        await asyncio.sleep(0)
        visit('class_amethod.2', cls)

    @classmethod
    @asyncSlot
    async def class_decorated_amethod(cls):
        visit('class_decorated_amethod.1', cls)
        await asyncio.sleep(0)
        visit('class_decorated_amethod.2', cls)

    # Not supported
    # @asyncSlot
    # @classmethod
    # async def class_decorated_amethod(cls):
    #     visit('class_decorated_amethod.1', cls)
    #     await asyncio.sleep(0)
    #     visit('class_decorated_amethod.2', cls)

    if qt_slot_supports_descriptor:

        @QtCore.Slot()
        @classmethod
        def slot_class_method(cls):
            visit('slot_class_method', cls)

        @classmethod
        @QtCore.Slot()
        def class_slot_method(cls):
            visit('class_slot_method', cls)


    # -------------------------------------------------------------------------
    # Static method
    # -------------------------------------------------------------------------

    @staticmethod
    def static_method():
        visit('static_method')

    @staticmethod
    async def static_amethod():
        visit('static_amethod.1')
        await asyncio.sleep(0)
        visit('static_amethod.2')

    @staticmethod
    @asyncSlot
    async def static_decorated_amethod():
        visit('static_decorated_amethod.1')
        await asyncio.sleep(0)
        visit('static_decorated_amethod.2')

    if qt_slot_supports_descriptor:

        @QtCore.Slot()
        @staticmethod
        def slot_static_method():
            visit('slot_static_method')

        @staticmethod
        @QtCore.Slot()
        def static_slot_method():
            visit('static_slot_method')


def func():
    visit('func')


@QtCore.Slot()
def slot_func():
    visit('slot_func')


async def afunc():
    visit('afunc.1')
    await asyncio.sleep(0)
    visit('afunc.2')


@asyncSlot
async def decorated_afunc():
    visit('decorated_afunc.1')
    await asyncio.sleep(0)
    visit('decorated_afunc.2')


@QtCore.Slot()
async def slot_afunc():
    visit('slot_afunc.1')
    await asyncio.sleep(0)
    visit('slot_afunc.2')


@QtCore.Slot()
@asyncSlot
async def slot_decorated_afunc():
    visit('slot_decorated_afunc.1')
    await asyncio.sleep(0)
    visit('slot_decorated_afunc.2')


@asyncSlot
@QtCore.Slot()
async def decorated_slot_afunc():
    visit('decorated_slot_afunc.1')
    await asyncio.sleep(0)
    visit('decorated_slot_afunc.2')


qc = QtCore.Qt.ConnectionType.QueuedConnection


class TestSlot(unittest.TestCase):

    def setUp(self) -> None:
        if QtCore.QCoreApplication.instance() is not None:
            self.app = QtCore.QCoreApplication.instance()
        else:
            self.app = QtCore.QCoreApplication([])

        self.qt_loop = QtCore.QEventLoop()
        self.receiver = MySlotObject()
        self.sender = MySignalObject()
        self.signal = self.sender.ready1
        self.connection = None

    def tearDown(self) -> None:
        if self.connection is not None:
            self.signal.disconnect()
            self.connection = None
        self.sender = None
        self.receiver = None
        self.qt_loop = None
        self.app = None

    def _run_once(self):
        QtCore.QTimer.singleShot(0, self.qt_loop.quit)
        self.signal.emit(True)

        called.clear()
        with AsyncSlotRunner():
            if hasattr(self.qt_loop, 'exec'):
                self.qt_loop.exec()
            else:
                self.qt_loop.exec_()
        return called.copy()

    # -------------------------------------------------------------------------
    # Test non-async free function without Qt.Slot decoration
    # -------------------------------------------------------------------------

    def test_func(self):
        # Qt sanity check
        self.connection = self.signal.connect(func, qc)
        result = self._run_once()
        self.assertEqual(result, ['func'])

    def test_wrapped_func(self):
        # Wrapping a func is an error
        with self.assertRaises(TypeError):
            asyncSlot(func)

    def test_decorated_func(self):
        # Decorating a func is an error
        with self.assertRaises(TypeError):
            @asyncSlot
            def fn():
                pass

    # -------------------------------------------------------------------------
    # Test non-async free function with Qt.Slot decoration
    # -------------------------------------------------------------------------

    def test_slot_func(self):
        # Qt sanity check
        self.connection = self.signal.connect(slot_func, qc)
        result = self._run_once()
        self.assertEqual(result, ['slot_func'])

    def test_wrapped_slot_func(self):
        # Wrapping a func is an error
        with self.assertRaises(TypeError):
            asyncSlot(slot_func)

    def test_decorated_slot_func(self):
        # Decorating a func is an error
        with self.assertRaises(TypeError):
            @asyncSlot
            @QtCore.Slot()
            def fn():
                pass

    # -------------------------------------------------------------------------
    # Test async free function without Qt.Slot decoration
    # -------------------------------------------------------------------------

    def test_wrapped_afunc(self):
        # Test wrapping afunc
        self.connection = self.signal.connect(asyncSlot(afunc), qc)
        result = self._run_once()
        self.assertEqual(result, ['afunc.1', 'afunc.2'])

    def test_decorated_afunc(self):
        # Test decorating afunc
        self.connection = self.signal.connect(decorated_afunc, qc)
        result = self._run_once()
        self.assertEqual(result, ['decorated_afunc.1', 'decorated_afunc.2'])

    def test_wrapped_decorated_afunc(self):
        # Wrapping a decorated afunc is an error
        with self.assertRaises(TypeError):
            asyncSlot(decorated_afunc)

    # -------------------------------------------------------------------------
    # Test async free function with Qt.Slot decoration
    # -------------------------------------------------------------------------

    def test_wrapped_slot_afunc(self):
        # Test wrapping slot_afunc
        self.connection = self.signal.connect(asyncSlot(slot_afunc), qc)
        result = self._run_once()
        self.assertEqual(result, ['slot_afunc.1', 'slot_afunc.2'])

    def test_decorated_slot_afunc(self):
        self.connection = self.signal.connect(decorated_slot_afunc, qc)
        result = self._run_once()
        self.assertEqual(result, ['decorated_slot_afunc.1',
                                  'decorated_slot_afunc.2'])

    def test_slot_decorated_afunc(self):
        self.connection = self.signal.connect(slot_decorated_afunc, qc)
        result = self._run_once()
        self.assertEqual(result, ['slot_decorated_afunc.1',
                                  'slot_decorated_afunc.2'])

    # -------------------------------------------------------------------------
    # Test non-async method
    # -------------------------------------------------------------------------

    def test_method(self):
        # Qt sanity check
        self.connection = self.signal.connect(self.receiver.method, qc)
        result = self._run_once()
        self.assertEqual(result, ['method(Self)'])

    def test_wrapped_method(self):
        # Wrapping a method is an error
        with self.assertRaises(TypeError):
            asyncSlot(self.receiver.method)

    def test_slot_method(self):
        # Qt sanity check
        self.connection = self.signal.connect(self.receiver.slot_method, qc)
        result = self._run_once()
        self.assertEqual(result, ['slot_method(Self)'])

    def test_wrapped_slot_method(self):
        # Wrapping a method is an error
        with self.assertRaises(TypeError):
            asyncSlot(self.receiver.slot_method)

    # -------------------------------------------------------------------------
    # Test async method
    # -------------------------------------------------------------------------

    def test_wrapped_amethod(self):
        self.connection = self.signal.connect(
            asyncSlot(self.receiver.amethod), qc)
        result = self._run_once()
        self.assertEqual(result, ['amethod.1(Self)', 'amethod.2(Self)'])

    def test_decorated_amethod(self):
        self.connection = self.signal.connect(
            self.receiver.decorated_amethod, qc)
        result = self._run_once()
        self.assertEqual(result, ['decorated_amethod.1(Self)',
                                  'decorated_amethod.2(Self)'])

    def test_wrapped_decorated_amethod(self):
        with self.assertRaises(TypeError):
            asyncSlot(self.receiver.decorated_amethod)

    def test_wrapped_slot_amethod(self):
        self.connection = self.signal.connect(
            asyncSlot(self.receiver.slot_amethod), qc)
        result = self._run_once()
        self.assertEqual(result, ['slot_amethod.1(Self)',
                                  'slot_amethod.2(Self)'])

    def test_decorated_slot_amethod(self):
        self.connection = self.signal.connect(
            self.receiver.decorated_slot_amethod, qc)
        result = self._run_once()
        self.assertEqual(result, ['decorated_slot_amethod.1(Self)',
                                  'decorated_slot_amethod.2(Self)'])

    def test_slot_decorated_amethod(self):
        self.connection = self.signal.connect(
            self.receiver.slot_decorated_amethod, qc)
        result = self._run_once()
        self.assertEqual(result, ['slot_decorated_amethod.1(Self)',
                                  'slot_decorated_amethod.2(Self)'])

    # -------------------------------------------------------------------------
    # Test non-async class method
    # -------------------------------------------------------------------------

    def test_class_method(self):
        # Qt sanity check
        self.connection = self.signal.connect(self.receiver.class_method, qc)
        result = self._run_once()
        self.assertEqual(result, ['class_method(Cls)'])

    def test_wrapped_class_method(self):
        with self.assertRaises(TypeError):
            asyncSlot(self.receiver.class_method)

    @unittest.skipUnless(qt_slot_supports_descriptor, 'not supported by PyQt')
    def test_slot_class_method(self):
        # Qt sanity check
        self.connection = self.signal.connect(
            self.receiver.slot_class_method, qc)
        result = self._run_once()
        self.assertEqual(result, ['slot_class_method(Cls)'])

    @unittest.skipUnless(qt_slot_supports_descriptor, 'not supported by PyQt')
    def test_class_slot_method(self):
        # Qt sanity check
        self.connection = self.signal.connect(
            self.receiver.class_slot_method, qc)
        result = self._run_once()
        self.assertEqual(result, ['class_slot_method(Cls)'])

    # -------------------------------------------------------------------------
    # Test async class method
    # -------------------------------------------------------------------------

    def test_wrapped_class_amethod(self):
        self.connection = self.signal.connect(
            asyncSlot(self.receiver.class_amethod), qc)
        result = self._run_once()
        self.assertEqual(result, ['class_amethod.1(Cls)',
                                  'class_amethod.2(Cls)'])

    def test_class_decorated_amethod(self):
        self.connection = self.signal.connect(
            self.receiver.class_decorated_amethod, qc)
        result = self._run_once()
        self.assertEqual(result, ['class_decorated_amethod.1(Cls)',
                                  'class_decorated_amethod.2(Cls)'])

    # -------------------------------------------------------------------------
    # Test non-async static method
    # -------------------------------------------------------------------------

    def test_static_method(self):
        # Qt sanity check
        self.connection = self.signal.connect(self.receiver.static_method, qc)
        result = self._run_once()
        self.assertEqual(result, ['static_method'])

    def test_wrapped_static_method(self):
        with self.assertRaises(TypeError):
            asyncSlot(self.receiver.static_method)

    @unittest.skipUnless(qt_slot_supports_descriptor, 'not supported by PyQt')
    def test_slot_static_method(self):
        # Qt sanity check
        self.connection = self.signal.connect(
            self.receiver.slot_static_method, qc)
        result = self._run_once()
        self.assertEqual(result, ['slot_static_method'])

    @unittest.skipUnless(qt_slot_supports_descriptor, 'not supported by PyQt')
    def test_static_slot_method(self):
        # Qt sanity check
        self.connection = self.signal.connect(
            self.receiver.static_slot_method, qc)
        result = self._run_once()
        self.assertEqual(result, ['static_slot_method'])

    # -------------------------------------------------------------------------
    # Test async static method
    # -------------------------------------------------------------------------

    def test_wrapped_static_amethod(self):
        self.connection = self.signal.connect(
            asyncSlot(self.receiver.static_amethod), qc)
        result = self._run_once()
        self.assertEqual(result, ['static_amethod.1', 'static_amethod.2'])

    def test_static_decorated_amethod(self):
        self.connection = self.signal.connect(
            self.receiver.static_decorated_amethod, qc)
        result = self._run_once()
        self.assertEqual(result, ['static_decorated_amethod.1',
                                  'static_decorated_amethod.2'])


if __name__ == '__main__':
    # TODO: insert sync callback to check invocation order
    unittest.main()



""" http_example.py - asynchronous http download and cancellation

This example displays a window that allows the user to download a web
page asynchronously and optionally cancel the download.

The window contains an 'Async GET' button whose 'clicked' signal is
connected to a coroutine function wrapped by asyncslot.asyncSlot.  The
window also contains a 'Cancel' button used to cancel the asynchronous
download operation.

To simulate slow network response so that it's easier to visualize the
difference between synchronous and asynchronous download, this example
implements a minimal http server that sleeps for 3 seconds before sending
back the response.  The http server runs in a separate thread and does
not use any of asyncslot's functionality.

Pay attention to code blocks marked with FEATURE -- they demonstrate the
usage of asyncslot.  The rest code builds up the GUI and program logic
and is not specific to asyncslot.
"""

import asyncio
import asyncslot
import sys
import time
from PyQt6 import QtWidgets
from bouncingwidget import BouncingWidget
from typing import Optional
import requests
import http.server
import threading
import httpx


def create_http_server():
    """ Create a local http server to simulate slow response """

    class MyRequestHandler(http.server.BaseHTTPRequestHandler):
        def do_GET(self):
            time.sleep(3)  # simulate slow response
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            content = f"You requested {self.path}".encode("utf-8")
            self.wfile.write(content)

    server = http.server.ThreadingHTTPServer(("127.0.0.1", 0),
                                             MyRequestHandler)
    thread = threading.Thread(target=server.serve_forever)
    thread.start()
    return server


class MyWidget(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("asyncslot - Http Example")

        # Show a bouncing ball to visualize whether the Qt event loop is
        # blocked -- the ball freezes if the Qt event loop is blocked.
        self._bouncer = BouncingWidget()

        # URl input.
        self._url = QtWidgets.QLineEdit(self)

        # The 'Sync GET' button downloads the web page synchronously,
        # which blocks the Qt event loop.  The bouncing ball freezes
        # when this button is clicked until download is complete.
        self._sync_button = QtWidgets.QPushButton("Sync GET")
        self._sync_button.clicked.connect(self.sync_download)

        # The 'Async GET' button downloads the web page asynchronously.
        # The ball keeps bouncing while the download is in progress.
        self._async_button = QtWidgets.QPushButton("Async GET")

        # ---- FEATURE ----
        # To connect an async function to the clicked signal, wrap the async
        # function in asyncslot.asyncSlot.
        self._async_button.clicked.connect(
            asyncslot.asyncSlot(self.async_download))

        # ---- FEATURE ----
        # When an async download is in progress, _async_task is set to the
        # task executing the download, so that the download may be cancelled
        # by clicking the 'Cancel' button.
        self._async_task: Optional[asyncio.Task] = None

        # The 'Cancel' button is enabled when async download is in progress.
        self._cancel_button = QtWidgets.QPushButton("Cancel")
        self._cancel_button.setEnabled(False)
        self._cancel_button.clicked.connect(self.cancel_async_download)

        # Response from the http server is shown in the below box.
        self._output = QtWidgets.QTextEdit(self)

        # Set up layout.
        self._buttons = QtWidgets.QHBoxLayout()
        self._buttons.addWidget(self._sync_button)
        self._buttons.addWidget(self._async_button)
        self._buttons.addWidget(self._cancel_button)
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.addWidget(self._bouncer)
        self._layout.addWidget(self._url)
        self._layout.addLayout(self._buttons)
        self._layout.addWidget(self._output)

        # Start a local HTTP server to simulate slow response.  The server
        # runs in a separate thread.
        self._server = create_http_server()

        # Set default URL to the locally-run http server, which sleeps for
        # 3 seconds before sending back a response.
        self._url.setText("http://{}:{}/dummy"
                          .format(*self._server.server_address))

    def closeEvent(self, event):
        # Handle the closeEvent to shut down the local HTTP server so that
        # the program exits (cleanly).  The shutdown() call blocks; you
        # may notice a short freeze of the bouncing ball.
        self._server.shutdown()
        event.accept()

    def sync_download(self):
        # When the 'Sync GET' button is clicked, download the web page
        # using the (blocking) requests library.  The bouncing ball freezes
        # until download is complete.  There is no need to disable buttons
        # etc because there is no chance for the Qt event loop to process
        # events or repaint the GUI.
        url = self._url.text() + "?sync"
        response = requests.get(url)
        self._output.setText(response.text)

    # ---- FEATURE ----
    # When the 'Async GET' button is clicked, initiate asynchronous download.
    async def async_download(self):
        # Retrieve and store the asyncio task wrapping this coroutine,
        # so that it may be cancelled by calling its cancel() method.
        self._async_task = asyncio.current_task()
        assert self._async_task is not None

        # Update GUI elements -- this is a common pattern in event handling.
        # Because asyncslot executes the slot immediately until the first
        # yield, the changes take effect immediately, eliminating potential
        # race conditions.  The actual GUI repainting happens after the first
        # yield, where control is given back to the Qt event loop.
        self._sync_button.setEnabled(False)
        self._async_button.setEnabled(False)
        self._cancel_button.setEnabled(True)
        self._output.clear()
        self._output.setEnabled(False)

        try:
            # Download web page using httpx library.  The httpx library
            # works with both asyncio and trio, and uses anyio to detect
            # the type of the running event loop.  That httpx works with
            # asyncslot shows that asyncslot's event loop behaves like
            # an asyncio event loop.
            async with httpx.AsyncClient() as client:
                url = self._url.text() + "?async"
                response = await client.get(url)
                # TODO: test asyncgen close
                body = await response.aread()
                self._output.setText(body.decode("utf-8"))

        except asyncio.CancelledError:
            # Catching a CancelledError indicates the task is cancelled.
            # This can happen either because the user clicked the 'Cancel'
            # button, or because the window is closed.  In the latter case
            # Qt's event loop exits and AsyncSlotRunner shuts down the
            # asyncslot event loop, cancelling all running tasks.
            if self.isVisible():
                # If the window is still open, the task must have been
                # cancelled by clicking the 'Cancel' button.  This is
                # because asyncslot only cancels running tasks _after_ the
                # Qt event loop exits, and the Qt event loop exits only
                # after all windows are closed.
                QtWidgets.QMessageBox.information(
                    self, "Note", "Download cancelled by user!")
            else:
                # Do nothing if user closed the window.
                pass

        finally:
            # Restore GUI element states.
            self._output.setEnabled(True)
            self._cancel_button.setEnabled(False)
            self._async_button.setEnabled(True)
            self._sync_button.setEnabled(True)
            self._async_task = None

    # ---- FEATURE ----
    # When the 'Cancel' button is clicked, cancel the async download.
    def cancel_async_download(self):
        # The 'Cancel' button is enabled only if the async download is
        # in progress, so the task object must be set.
        assert self._async_task is not None

        # Initiate cancellation request.  This throws asyncio.CancelledError
        # into the (suspended) coroutine, which must catch the exception and
        # perform actual cancellation.  (Note that it is possible for the
        # task to be done before CancelledError is thrown, as a cancellation
        # request is scheduled by call_soon and it could happen that
        # a task completion callback is scheduled before it.)  Cancelling
        # a done task has no effect.
        self._async_task.cancel()


def main():
    # Create a QApplication instance, as usual.
    app = QtWidgets.QApplication([])

    # Create widgets, as usual.
    widget = MyWidget()
    widget.resize(400, 300)
    widget.show()

    # ---- FEATURE ----
    # For asyncslot to work, enclose app.exec inside the context manager
    # AsyncSlotRunner().  The runner is responsible for starting up and
    # shutting down the logical asyncio event loop.
    with asyncslot.AsyncSlotRunner():
        sys.exit(app.exec())


if __name__ == "__main__":
    main()



""" bouncingwidget.py - Qt widget that displays a bouncing ball """

from PyQt6 import QtCore, QtWidgets, QtGui


__all__ = ("BouncingWidget", )


class BouncingWidget(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self._x = 0
        self._y = 0
        self._d = 20
        self._vx = 10
        self._vy = 5
        self.setMinimumSize(self._d * 2, self._d * 2)
        self.startTimer(50)

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        w = self.width()
        h = self.height()
        self._x += self._vx
        if self._x < 0:
            self._x = -self._x
            self._vx = -self._vx
        elif self._x + self._d > w:
            self._x -= (self._x + self._d - w)
            self._vx = -self._vx
        self._y += self._vy
        if self._y < 0:
            self._y = -self._y
            self._vy = -self._vy
        elif self._y + self._d > h:
            self._y -= (self._y + self._d - h)
            self._vy = -self._vy
        self.update()

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.drawRect(0, 0, self.width(), self.height())
        brush = QtGui.QBrush(QtCore.Qt.GlobalColor.black)
        painter.setBrush(brush)
        painter.drawEllipse(self._x, self._y, self._d, self._d)



""" _windows_events.py - implements proactor event loop under Windows """

import sys

if sys.platform != 'win32':
    raise ImportError('win32 only')

import _overlapped
import _winapi
import asyncio.windows_events
import concurrent.futures
import math
import sys
import threading
from typing import Optional
from ._selectable import _AsyncSlotNotifier
from . import _proactor_events
from . import _selector_events


__all__ = (
    'AsyncSlotDefaultEventLoop',
    'AsyncSlotDefaultEventLoopPolicy',
    'AsyncSlotProactorEventLoop',
    'AsyncSlotProactorEventLoopPolicy',
    'AsyncSlotSelectorEventLoop',
    'AsyncSlotSelectorEventLoopPolicy',
)


INFINITE = 0xffffffff


class _AsyncSlotProactor(asyncio.IocpProactor):
    def __init__(self, concurrency=0xffffffff):
        super().__init__(concurrency)

        self.__executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self.__dequeue_future: Optional[concurrent.futures.Future] = None
        self.__idle = threading.Event()
        self.__idle.set()
        self.__notifier: Optional[_AsyncSlotNotifier] = None

    def __wakeup(self):
        self._check_closed()
        if not self.__idle.is_set():
            assert self.__notifier is not None, 'notifier expected'
            self.__notifier.wakeup()
            self.__idle.wait()

    def set_notifier(self, notifier: Optional[_AsyncSlotNotifier]) -> None:
        self.__wakeup()
        self.__notifier = notifier

    def _poll(self, timeout=None):
        # _poll is called by super().select() and super().close().
        #
        # If the last call to select() raised AsyncSlotYield, the caller
        # (from _run_once) should only call select() again after receiving
        # a notification from us, and we only send a notification after
        # entering IDLE state.
        #
        # If called from close(), we also require the proactor to have
        # been woken up (by a call to set_notifier) before closing.
        #
        # The code below is copied verbatim from asyncio.windows_events,
        # except that _overlapped is 'redirected' to this object's
        # non-blocking implementation.  The code is unchanged from Python
        # 3.7 through Python 3.11.
        _overlapped = self

        # --- BEGIN COPIED FROM asyncio.windows_events.IocpProactor._poll
        if timeout is None:
            ms = INFINITE
        elif timeout < 0:
            raise ValueError("negative timeout")
        else:
            # GetQueuedCompletionStatus() has a resolution of 1 millisecond,
            # round away from zero to wait *at least* timeout seconds.
            ms = math.ceil(timeout * 1e3)
            if ms >= INFINITE:
                raise ValueError("timeout too big")

        while True:
            status = _overlapped.GetQueuedCompletionStatus(self._iocp, ms)
            if status is None:
                break
            ms = 0

            err, transferred, key, address = status
            try:
                f, ov, obj, callback = self._cache.pop(address)
            except KeyError:
                if self._loop.get_debug():
                    self._loop.call_exception_handler({
                        'message': ('GetQueuedCompletionStatus() returned an '
                                    'unexpected event'),
                        'status': ('err=%s transferred=%s key=%#x address=%#x'
                                   % (err, transferred, key, address)),
                    })

                # key is either zero, or it is used to return a pipe
                # handle which should be closed to avoid a leak.
                if key not in (0, _overlapped.INVALID_HANDLE_VALUE):
                    _winapi.CloseHandle(key)
                continue

            if obj in self._stopped_serving:
                f.cancel()
            # Don't call the callback if _register() already read the result or
            # if the overlapped has been cancelled
            elif not f.done():
                try:
                    value = callback(transferred, key, ov)
                except OSError as e:
                    f.set_exception(e)
                    self._results.append(f)
                else:
                    f.set_result(value)
                    self._results.append(f)

        # Remove unregistered futures
        for ov in self._unregistered:
            self._cache.pop(ov.address, None)
        self._unregistered.clear()
        # --- END COPIED FROM asyncio.windows_events.IocpProactor._poll

    def GetQueuedCompletionStatus(self, iocp, ms):
        assert iocp is self._iocp

        assert self.__idle.is_set(), 'unexpected _poll'

        # If any prior dequeue result is available, return that.
        if self.__dequeue_future is not None:
            try:
                return self.__dequeue_future.result()
            finally:
                self.__dequeue_future = None

        # Perform normal (blocking) polling if no notifier is set.
        # In particular, this is the case when called by close().
        if self.__notifier is None:
            return _overlapped.GetQueuedCompletionStatus(self._iocp, ms)

        # Perform normal polling if timeout is zero.
        if ms == 0:
            return _overlapped.GetQueuedCompletionStatus(self._iocp, ms)

        # Try non-blocking dequeue and return if any result is available.
        status = _overlapped.GetQueuedCompletionStatus(self._iocp, 0)
        if status is not None:
            return status

        # Launch a thread worker to wait for IO.
        self.__idle.clear()
        try:
            self.__dequeue_future = self.__executor.submit(self.__dequeue, ms)
        except BaseException:
            # Should submit() raise, we assume no task is spawned.
            self.__idle.set()
            raise
        else:
            return self.__notifier.no_result()  # raises _AsyncSlotYield

    def __dequeue(self, ms: int):
        try:
            # Note: any exception raised is propagated to the main thread
            # the next time select() is called, and will bring down the
            # AsyncSlotEventLoop.
            return _overlapped.GetQueuedCompletionStatus(self._iocp, ms)
        finally:
            # Make a copy of self.__notifier because it may be altered by
            # set_notifier immediately after self.__idle is set.
            notifier = self.__notifier
            self.__idle.set()
            notifier.notify()

    def close(self):
        assert self.__idle.is_set(), 'unexpected close'
        assert self.__notifier is None, 'notifier must have been reset'

        # Note: super().close() calls self._poll() repeatedly to exhaust
        # IO events.  The first call might be served by __dequeue_future;
        # the remaining calls are guaranteed to block because __notifier
        # is None.
        super().close()

        if self.__executor is not None:
            self.__executor.shutdown()
            self.__executor = None


class AsyncSlotProactorEventLoop(
    _proactor_events.AsyncSlotBaseProactorEventLoop,
    asyncio.windows_events.ProactorEventLoop
):
    def __init__(self):
        proactor = _AsyncSlotProactor()
        super().__init__(proactor)

    if sys.version_info >= (3, 8):
        # run_forever is overridden in Python 3.8 and above

        def _asyncslot_loop_startup(self):
            # ---- BEGIN COPIED FROM ProactorEventLoop.run_forever
            assert self._self_reading_future is None
            self.call_soon(self._loop_self_reading)
            # ---- END COPIED FROM ProactorEventLoop.run_forever
            super()._asyncslot_loop_startup()

        def _asyncslot_loop_cleanup(self):
            super()._asyncslot_loop_cleanup()
            # ---- BEGIN COPIED FROM ProactorEventLoop.run_forever
            if self._self_reading_future is not None:
                ov = self._self_reading_future._ov
                self._self_reading_future.cancel()
                # self_reading_future was just cancelled so if it hasn't been
                # finished yet, it never will be (it's possible that it has
                # already finished and its callback is waiting in the queue,
                # where it could still happen if the event loop is restarted).
                # Unregister it otherwise IocpProactor.close will wait for it
                # forever
                if ov is not None:
                    self._proactor._unregister(ov)
                self._self_reading_future = None
            # ---- END COPIED FROM ProactorEventLoop.run_forever


class AsyncSlotProactorEventLoopPolicy(
    asyncio.events.BaseDefaultEventLoopPolicy
):
    _loop_factory = AsyncSlotProactorEventLoop


class AsyncSlotSelectorEventLoop(
    _selector_events.AsyncSlotBaseSelectorEventLoop,
    asyncio.windows_events.SelectorEventLoop
):
    pass


class AsyncSlotSelectorEventLoopPolicy(
    asyncio.events.BaseDefaultEventLoopPolicy
):
    _loop_factory = AsyncSlotSelectorEventLoop


if sys.version_info < (3, 8):
    AsyncSlotDefaultEventLoop = AsyncSlotSelectorEventLoop
    AsyncSlotDefaultEventLoopPolicy = AsyncSlotSelectorEventLoopPolicy
else:
    AsyncSlotDefaultEventLoop = AsyncSlotProactorEventLoop
    AsyncSlotDefaultEventLoopPolicy = AsyncSlotProactorEventLoopPolicy



""" _base_events.py - event loop implementation using Qt """

import asyncio
import signal
import sys
import threading
import traceback
from asyncio import events
from typing import Optional
from ._selectable import *


__all__ = 'AsyncSlotBaseEventLoop',


class _AsyncSlotYield(Exception):
    """ Raised by an AsyncSlotSelectable to indicate that no IO is readily
    available and that _run_once should yield to the Qt event loop. """
    pass


class _InterruptEvent:
    __slots__ = '_flag',

    def __init__(self):
        self._flag = False

    def set(self) -> None:
        self._flag = True

    def is_set(self) -> bool:
        return self._flag

    def clear(self) -> None:
        self._flag = False


def _interrupt_handler(sig, frame):
    assert sig == signal.SIGINT
    # if frame is not None:
    #     print(frame.f_locals)
    if frame is not None and '_interrupt_event' in frame.f_locals:
        _interrupt_event = frame.f_locals['_interrupt_event']
        if isinstance(_interrupt_event, _InterruptEvent):
            _interrupt_event.set()
            return
    return signal.default_int_handler(sig, frame)


_AsyncSlotNotifierObject = None


def _create_notifier(loop: "AsyncSlotBaseEventLoop"):
    global _AsyncSlotNotifierObject
    if _AsyncSlotNotifierObject is not None:
        return _AsyncSlotNotifierObject(loop)

    from .bindings import QtCore

    class _AsyncSlotNotifierObject(QtCore.QObject):
        if hasattr(QtCore, "pyqtSignal"):
            _notified = QtCore.pyqtSignal()
        else:
            _notified = QtCore.Signal()

        def __init__(self, loop: "AsyncSlotBaseEventLoop"):
            super().__init__()
            # The following creates a reference cycle.  Call close() to
            # break the cycle.
            self._loop = loop
            # Must make queued connection to avoid re-entrance.
            self._notified.connect(self._on_notified,
                                   QtCore.Qt.ConnectionType.QueuedConnection)
            # Install a SIGINT handler if no custom handler is installed.
            self._interrupt_handler_installed = False
            if signal.getsignal(signal.SIGINT) is signal.default_int_handler:
                try:
                    signal.signal(signal.SIGINT, _interrupt_handler)
                except (ValueError, OSError):
                    pass
                else:
                    self._interrupt_handler_installed = True

        def _on_notified(self, _interrupt_event=_InterruptEvent()):
            # If Ctrl+C is pressed while the loop is in a 'non-blocking'
            # select(), the select() will be woken up (due to set_wakeup_fd)
            # and the _notified signal emitted.  KeyboardInterrupt will be
            # raised at the first point where Python byte code is run, i.e.
            # this method.  We wrap the body in try-except to handle this.
            try:
                if _interrupt_event.is_set():
                    raise KeyboardInterrupt
                if self._loop is not None:
                    self._loop._asyncslot_loop_iteration()
                else:
                    # TODO: print a warning that notification is received after
                    # TODO: the notifier is closed.
                    pass
            except KeyboardInterrupt as exc:
                # We catch Ctrl+C only once.  If Ctrl+C is pressed again
                # immediately, the program will crash.
                if self._loop is not None:
                    self._loop._asyncslot_loop_interrupt(exc)
                else:
                    pass  # ignore Ctrl+C for once
            finally:
                _interrupt_event.clear()

        def no_result(self):
            raise _AsyncSlotYield

        def notify(self):
            self._notified.emit()

        def wakeup(self):
            self._loop._write_to_self()

        def close(self):
            if self._loop is not None:
                self._notified.disconnect()
                self._loop = None
            if self._interrupt_handler_installed:
                try:
                    if signal.getsignal(signal.SIGINT) is _interrupt_handler:
                        signal.signal(signal.SIGINT, signal.default_int_handler)
                except (ValueError, OSError):
                    pass
                finally:
                    self._interrupt_handler_installed = False

    return _AsyncSlotNotifierObject(loop)


class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
    """Implements the scheduling logic of asyncslot event loop.

    An asyncio event loop can be in one of the following 'states':
      - STOPPED: loop is not polling for IO or processing events
      - RUNNING: loop is polling for IO or processing events
      - CLOSED: loop is closed and cannot be used in any way

    The _closed and _thread_id attributes determine the current state:

                 _closed    _thread_id
      RUNNING    False      not None
      STOPPED    False      None
      CLOSED     True       None

    A RUNNING AsyncSlotEventLoop can run in one of the following 'modes':
      - STANDALONE: run_forever creates a QEventLoop and calls exec()
      - INTEGRATED: user is responsible for creating a Qt event loop
      - EXCLUSIVE:  run_forever creates a true asyncio event loop

    The __qt_event_loop and __notifier attributes determine the current
    running mode (provided the loop is RUNNING):

                    __qt_event_loop    __notifier
      STANDALONE    not None           not None
      INTEGRATED    None               not None
      EXCLUSIVE     None               None

    A RUNNING AsyncSlotEventLoop in STANDALONE or INTEGRATED mode can
    further be in one of two sub-states: PROCESSING or SELECTING.

      - PROCESSING: __process_asyncio_events() is being executed
      - SELECTING: __process_asyncio_events() is waiting to be scheduled

    A RUNNING AsyncSlotEventLoop in EXCLUSIVE mode is always PROCESSING.

    The __processing attribute determines the current sub-state (provided
    the loop is RUNNING).

    STANDALONE Mode
    ---------------

    A loop created with standalone == True launches a QEventLoop and calls
    exec() on it when run_forever is called.  If no QObjects are accessed,
    such a loop provides 100% API and semantics compatibility with asyncio.
    This mode is mainly used for testing the conformance of asyncslot's
    loop implementation.

    INTEGRATED Mode & EXCLUSIVE Mode
    --------------------------------

    A loop created with standalone == False never launches a Qt event loop
    itself.  The normal workflow is for the user to call start() first and
    then launch a Qt event loop, e.g. by calling one of QEventLoop.exec(),
    QCoreApplication.exec(), QThread.exec(), etc.  After the Qt event loop
    exits, the user should call stop().  Such a loop run is said to be in
    INTEGRATED mode.

    Note that start() and stop() only change the (logical) state of an
    AsyncSlotEventLoop; they have no effect on the (physical) Qt event
    loop and is in fact independent of the lifetime of the latter.

    If run_forever() is called on a loop created with standalone == False,
    a 'true' asyncio event loop is run, which blocks the calling thread
    until stop() is called.  This is known as EXCLUSIVE mode, and is
    designed for running clean-up code after the Qt event loop has exited.
    Such clean-up code should finish as soon as possible, and should not
    access any Qt object as no Qt loop is running.
    """
    def __init__(self, *args, **kwargs):
        # If True, run_forever will launch the loop in STANDALONE mode.
        # If False, run_forever will launch the loop in EXCLUSIVE mode;
        # use start() to launch the loop in INTEGRATED mode.
        self.__standalone = True

        # If self is created in STANDALONE mode and is running,
        # __qt_event_loop is set to the QEventLoop that is being run.
        # In any other case, __qt_event_loop is set to None.
        self.__qt_event_loop = None

        # If self is running in STANDALONE or INTEGRATED mode, __notifier
        # is set to a notifier object and is attached to the selector to
        # notify us when IO is available or timeout occurs.  In any other
        # case, __notifier is set to None.
        self.__notifier: Optional[_AsyncSlotNotifier] = None

        # __processing is set to True in _asyncslot_loop_iteration to
        # indicate that a 'normal' asyncio event processing iteration
        # (i.e. _run_once) is running.  It is also set to True when the
        # loop is running in EXCLUSIVE mode.
        self.__processing = False

        # Any exception raised by self._process_asyncio_events is stored
        # in __run_once_error to be propagated later to the caller of
        # self.run_forever, as QEventLoop.exec() does not propagate
        # exceptions.  Exceptions raised by tasks are normally not
        # propagated except for SystemExit and KeyboardInterrupt.
        self.__run_once_error: Optional[BaseException] = None

        # When the loop is running in INTEGRATED or STANDALONE mode,
        # __old_agen_hook is set to the asyncgen hooks before the loop is
        # started, so that they can be restored after the loop is stopped.
        self.__old_agen_hooks: Optional[tuple] = None

        # If __call_soon_eagerly is True, _call_soon does not schedule the
        # callback but instead invoke it immediately.  This flag is used by
        # run_task to eagerly execute the first step of a task.
        self.__call_soon_eagerly = False

        # Need to invoke base constructor after initializing member variables
        # for compatibility with Python 3.7's BaseProactorEventLoop (Windows),
        # which calls self.call_soon() indirectly from its constructor.
        super().__init__(*args, **kwargs)  # noqa

    # =========================================================================
    # Custom method for AsyncSlot
    # =========================================================================

    def set_guest(self, guest: bool) -> None:
        self._check_closed()
        self._check_running()
        self.__standalone = not guest

    def run_task(self, coro, *, name=None):
        try:
            self.__call_soon_eagerly = True
            if name is None:
                return self.create_task(coro)
            else:
                return self.create_task(coro, name=name)
        finally:
            # This flag is normally reset by _call_soon, but also reset here
            # in case _call_soon is not called due to exception.
            self.__call_soon_eagerly = False

    def start(self) -> None:
        if self.__standalone:
            raise RuntimeError('AsyncSlotBaseEventLoop.start() can only be '
                               'called for a loop operating in guest mode')
        self._asyncslot_loop_startup()

    def _asyncslot_loop_startup(self) -> None:
        """ Start the logical asyncio event loop. """

        # ---- BEGIN COPIED FROM BaseEventLoop.run_forever
        self._check_closed()
        self._check_running()
        self._set_coroutine_origin_tracking(self._debug)
        self._thread_id = threading.get_ident()

        old_agen_hooks = sys.get_asyncgen_hooks()
        sys.set_asyncgen_hooks(firstiter=self._asyncgen_firstiter_hook,
                               finalizer=self._asyncgen_finalizer_hook)
        # ---- END COPIED FROM BaseEventLoop.run_forever

        self.__old_agen_hooks = old_agen_hooks

        self.__notifier = _create_notifier(self)
        self.__notifier.notify()  # schedule initial _run_once

        # Do not set notifier if a TestSelector is used (during testing).
        if hasattr(self._selector, 'set_notifier'):
            self._selector.set_notifier(self.__notifier)  # noqa

        events._set_running_loop(self)  # TODO: what does this do?

    def _asyncslot_loop_cleanup(self) -> None:
        """ Stop the logical asyncio event loop. """
        if self.__old_agen_hooks is not None:
            old_agen_hooks = self.__old_agen_hooks
            self.__old_agen_hooks = None
        else:
            old_agen_hooks = sys.get_asyncgen_hooks()
        if self.__notifier is not None:
            # Do not set notifier if a TestSelector is used (during testing).
            if hasattr(self._selector, 'set_notifier'):
                self._selector.set_notifier(None)  # noqa
            self.__notifier.close()
            self.__notifier = None
        # ---- BEGIN COPIED FROM BaseEventLoop.run_forever
        self._stopping = False
        self._thread_id = None
        events._set_running_loop(None)
        self._set_coroutine_origin_tracking(False)
        sys.set_asyncgen_hooks(*old_agen_hooks)
        # ---- END COPIED FROM BaseEventLoop.run_forever

    def _asyncslot_loop_iteration(self):
        """ This slot is connected to the notified signal of self.__notifier,
        which is emitted whenever asyncio events are possibly available
        and need to be processed."""
        assert not self.is_closed(), 'loop unexpectedly closed'
        assert self.is_running(), 'loop unexpectedly stopped'

        # Process ready callbacks, ready IO, and scheduled callbacks that
        # have passed the schedule time.  Run only once to avoid starving
        # the Qt event loop.
        try:
            self.__processing = True
            self._run_once()  # defined in asyncio.BaseEventLoop
        except _AsyncSlotYield:
            # Ignore _stopping flag until select() returns.  This follows
            # asyncio behavior.
            pass
        except BaseException as exc:
            self._asyncslot_loop_interrupt(exc)
        else:
            # To be consistent with asyncio behavior, check the _stopping
            # flag only after running a full iteration of _run_once.
            if self._stopping:
                if self.__qt_event_loop is not None:
                    # In STANDALONE mode, quit the loop and let run_forever
                    # perform clean up.
                    self.__qt_event_loop.exit(0)
                else:
                    # In INTEGRATED mode, stop immediately because there is
                    # no 'caller' to perform the cleanup for us.
                    self._asyncslot_loop_cleanup()
            else:
                # Schedule next iteration if this iteration did not block
                self.__notifier.notify()
        finally:
            self.__processing = False

    def _asyncslot_loop_interrupt(self, exc: BaseException):
        """Terminate the loop abnormally with the given exception.

        This method is called by:

          (1) AsyncSlotBaseEventLoop._asyncslot_loop_iteration() if
              _run_once raised a BaseException, typically SystemExit
              or KeyboardInterrupt; and

          (2) _AsyncSlotNotifier._on_notified() if KeyboardInterrupt
              is raised during processing the notification.

        """
        if self.__qt_event_loop is not None:
            # In STANDALONE mode, propagate the exception to the caller
            # of run_forever.
            self.__run_once_error = exc
            self.__qt_event_loop.exit(1)
        else:
            # In INTEGRATED mode, stop the loop immediately and raise
            # the error into the Qt event loop.  For PyQt this will
            # terminate the process; for PySide this will log an error
            # and then ignored.
            # TODO: if KeyboardInterrupt is raised, the object may not
            # TODO: be in a consistent state to perform cleanup.
            self._asyncslot_loop_cleanup()
            raise exc
            # TODO: implement consistent behavior under both bindings.
            # TODO: maybe call some exception handler and let it decide
            # TODO: what to do?

    # =========================================================================
    # Compatibility with Python 3.7
    # =========================================================================

    if sys.version_info < (3, 7):
        raise RuntimeError('asyncslot requires Python 3.7 or higher')

    elif sys.version_info < (3, 8):
        _check_running = asyncio.BaseEventLoop._check_runnung

    # =========================================================================
    # Methods defined in asyncio.AbstractEventLoop
    # =========================================================================

    # -------------------------------------------------------------------------
    # Running and stopping the event loop.
    # -------------------------------------------------------------------------

    def run_forever(self) -> None:
        """ Run the event loop until stop() is called. """

        # If run_forever is called in INTEGRATED mode, start an 'authentic'
        # asyncio event loop (by leaving __notifier to None).
        if not self.__standalone:
            try:
                self.__processing = True
                return super().run_forever()
            finally:
                self.__processing = False

        from .bindings import QtCore
        if QtCore.QCoreApplication.instance() is None:
            # TODO: do we need the same check in start()?
            raise RuntimeError('An instance of QCoreApplication or its '
                               'derived class must be create before running '
                               'AsyncSlotEventLoop')

        try:
            self._asyncslot_loop_startup()
            self.__qt_event_loop = QtCore.QEventLoop()
            if hasattr(QtCore.QEventLoop, 'exec'):
                exit_code = self.__qt_event_loop.exec()
            else:
                exit_code = self.__qt_event_loop.exec_()
            if exit_code != 0:
                # Propagate exception from _process_asyncio_events if
                # one is set.  The exception is not set if the Qt loop
                # is terminated by e.g. QCoreApplication.exit().  Note
                # also that if QCoreApplication.exit() has been called
                # before calling run_forever, QEventLoop.exec() would
                # return immediately.
                if self.__run_once_error is not None:
                    raise self.__run_once_error  # TODO: test this
                else:
                    raise RuntimeError(
                        f"Qt event loop exited with code '{exit_code}'")
        except BaseException:
            # TODO: improve diagnosis
            print(traceback.format_exc(), file=sys.stderr)
            raise
        finally:
            self.__run_once_error = None
            self.__qt_event_loop = None
            self._asyncslot_loop_cleanup()

    # run_until_complete = BaseEventLoop.run_until_complete

    def stop(self) -> None:
        """ Request the loop to stop.

        The precise semantics are as follows:

        If the loop is operating in 'host' mode:

          - If stop() is called when the loop is STOPPED, the next time the
            loop is RUNNING it will run exactly one iteration and then stop.

          - If stop() is called from a callback of a RUNNING loop, the loop
            will stop after completing the current iteration.

          - If stop() is called from interleaving code (a Qt slot) when the
            loop is RUNNING, treat as if called via call_soon_threadsafe():
            wake up the selector, which will run one iteration and stop.

        If the loop is operating in 'guest' mode:

          - If stop() is called when the loop is STOPPED, raise an error.
            (The pre-stop idiom is not supported.)

          - If stop() is called from a callback of a RUNNING loop, either
            in INTEGRATED or EXCLUSIVE mode, the loop will stop after
            completing the current iteration.

          - If stop() is called from interleaving code (a Qt slot) when
            the loop is RUNNING in INTEGRATED mode, wake up the selector
            and stop the loop immediately, since there might not be a Qt
            loop to execute the next iteration.

        In any case, if KeyboardInterrupt or SystemExit is raised when
        processing an iteration, that iteration is interrupted, the loop
        is stopped immediately, and the exception is propagated to the
        caller of run_forever if operating in 'host' mode or thrown into
        the Qt loop if operating in 'guest' mode.
        """
        if self.__standalone:
            if self.is_running() and not self.__processing:
                self._write_to_self()
            super().stop()

        elif not self.is_running():
            raise RuntimeError('AsyncSlotEventLoop: stop can only be called '
                               'when a loop created in integrated mode is '
                               'running')

        elif self.__processing:
            super().stop()

        else:
            self._write_to_self()
            super().stop()  # this only sets the flat
            self._asyncslot_loop_cleanup()

    # is_running = BaseEventLoop.is_running
    # is_closed = BaseEventLoop.is_closed
    # close = BaseEventLoop.close
    # shutdown_asyncgens = BaseEventLoop.shutdown_asyncgens
    # shutdown_default_executor = BaseEventLoop.shutdown_default_executor

    # -------------------------------------------------------------------------
    # Methods scheduling callbacks.  All these return Handles.
    # -------------------------------------------------------------------------

    # _timer_handle_cancelled: see BaseEventLoop

    def call_soon(self, *args, **kwargs):
        # If called from interleaving code when the loop is SELECTING,
        # treat as if called by call_soon_threadsafe().
        if self.is_running() and not self.__processing:
            self._write_to_self()

        # Eager execution if called from run_task().
        handle = super().call_soon(*args, **kwargs)
        if self.__call_soon_eagerly:
            self.__call_soon_eagerly = False
            # asyncio does not support recursive task execution, so 'suspend'
            # the current task before running the child task and 'resume' it
            # after the child task completes one step.
            current_task = asyncio.tasks.current_task(self)
            if current_task is not None:
                asyncio.tasks._leave_task(self, current_task)
            try:
                # only propagates SystemExit and KeyboardInterrupt
                handle._run()
            finally:
                # Cancel the handle because it is already in the _ready queue
                handle.cancel()
                # Resume the parent task if any.
                if current_task is not None:
                    asyncio.tasks._enter_task(self, current_task)
        return handle

    def call_later(self, *args, **kwargs):
        if self.is_running() and not self.__processing:
            self._write_to_self()
        return super().call_later(*args, **kwargs)

    def call_at(self, *args, **kwargs):
        if self.is_running() and not self.__processing:
            self._write_to_self()
        return super().call_at(*args, **kwargs)

    # time: see BaseEventLoop
    # create_future: see BaseEventLoop

    # -------------------------------------------------------------------------
    # Method scheduling a coroutine object: create a task.
    # -------------------------------------------------------------------------

    # create_task = BaseEventLoop.create_task

    # -------------------------------------------------------------------------
    # Methods for interacting with threads.
    # -------------------------------------------------------------------------

    # call_soon_threadsafe: BaseEventLoop
    # run_in_executor: BaseEventLoop
    # set_default_executor: BaseEventLoop

    # -------------------------------------------------------------------------
    # Network I/O methods returning Futures.
    # -------------------------------------------------------------------------

    # getaddrinfo = BaseEventLoop.getaddrinfo
    # getnameinfo = BaseEventLoop.getnameinfo
    # create_connection = BaseEventLoop.create_connection
    # create_server = BaseEventLoop.create_server
    # sendfile = BaseEventLoop.sendfile
    # start_tls = BaseEventLoop.start_tls
    # create_unix_connection = _UnixSelectorEventLoop.create_unix_connection
    # create_unix_server = _UnixSelectorEventLoop.create_unix_server
    # create_datagram_endpoint = BaseEventLoop.create_datagram_endpoint

    # -------------------------------------------------------------------------
    # Pipes and subprocesses.
    # -------------------------------------------------------------------------

    # connect_read_pipe = BaseEventLoop.connect_read_pipe
    # connect_write_pipe = BaseEventLoop.connect_write_pipe
    # subprocess_shell = BaseEventLoop.subprocess_shell
    # subprocess_exec = BaseEventLoop.subprocess_exec

    # -------------------------------------------------------------------------
    # Ready-based callback registration methods.
    # -------------------------------------------------------------------------

    # add_reader = BaseSelectorEventLoop.add_reader
    # remove_reader = BaseSelectorEventLoop.remove_reader
    # add_writer = BaseSelectorEventLoop.add_writer
    # remove_writer = BaseSelectorEventLoop.remove_writer

    # -------------------------------------------------------------------------
    # Completion based I/O methods returning Futures.
    # -------------------------------------------------------------------------

    # sock_recv = BaseSelectorEventLoop.sock_recv
    # sock_recv_into = BaseSelectorEventLoop.sock_recv_into
    # sock_sendall = BaseSelectorEventLoop.sock_sendall
    # sock_connect = BaseSelectorEventLoop.sock_connect
    # sock_accept = BaseSelectorEventLoop.sock_accept
    # sock_sendfile = BaseSelectorEventLoop.sock_sendfile

    # -------------------------------------------------------------------------
    # Signal handling.
    # -------------------------------------------------------------------------

    # add_signal_handler = _UnixSelectorEventLoop.add_signal_handler
    # remove_signal_handler = _UnixSelectorEventLoop.remove_signal_handler

    # -------------------------------------------------------------------------
    # Task factory.
    # -------------------------------------------------------------------------

    # set_task_factory = BaseEventLoop.set_task_factory
    # get_task_factory = BaseEventLoop.get_task_factory

    # -------------------------------------------------------------------------
    # Error handlers.
    # -------------------------------------------------------------------------

    # get_exception_handler = BaseEventLoop.get_exception_handler
    # set_exception_handler = BaseEventLoop.set_exception_handler
    # default_exception_handler = BaseEventLoop.default_exception_handler
    # call_exception_handler = BaseEventLoop.call_exception_handler

    # -------------------------------------------------------------------------
    # Debug flag management.
    # -------------------------------------------------------------------------

    # get_debug = BaseEventLoop.get_debug
    # set_debug = BaseEventLoop.set_debug

    # =========================================================================
    # Abstract methods defined by asyncio.BaseEventLoop
    # =========================================================================

    # _make_socket_transport = BaseSelectorEventLoop._make_socket_transport
    # _make_ssl_transport = BaseSelectorEventLoop._make_ssl_transport
    # _make_datagram_transport = BaseSelectorEventLoop._make_datagram_transport
    # _make_read_pipe_transport: see _UnixSelectorEventLoop
    # _make_write_pipe_transport: see _UnixSelectorEventLoop
    # _make_subprocess_transport: see _UnixSelectorEventLoop
    # _write_to_self: see BaseSelectorEventLoop / BaseProactorEventLoop
    # _process_events: see BaseSelectorEventLoop / BaseProactorEventLoop



""" _unix_events.py - define default loop and policy under unix """

import sys

if sys.platform == 'win32':
    raise ImportError('unix only')


import asyncio.unix_events
from . import _selector_events


__all__ = (
    'AsyncSlotDefaultEventLoop',
    'AsyncSlotDefaultEventLoopPolicy',
    'AsyncSlotSelectorEventLoop',
    'AsyncSlotSelectorEventLoopPolicy',
)


class AsyncSlotSelectorEventLoop(
    _selector_events.AsyncSlotBaseSelectorEventLoop,
    asyncio.unix_events.SelectorEventLoop
):
    def remove_signal_handler(self, sig):
        result = super().remove_signal_handler(sig)
        if not self._signal_handlers and not self._closed:
            # AsyncSlotBaseSelectorEventLoop installs a wakeup fd, but
            # _UnixSelectorEventLoop.remove_signal_handler uninstalls
            # it if there are no signal handlers.  This is not what we
            # want.  Re-install the wakeup in this case.
            self._asyncslot_install_wakeup_fd()
        return result


class AsyncSlotSelectorEventLoopPolicy(
    asyncio.unix_events.DefaultEventLoopPolicy
):
    _loop_factory = AsyncSlotSelectorEventLoop


AsyncSlotDefaultEventLoopPolicy = AsyncSlotSelectorEventLoopPolicy
AsyncSlotDefaultEventLoop = AsyncSlotSelectorEventLoop



""" _proactor_events.py - no-op counterpart to asyncio.BaseProactorEventLoop """

import asyncio.proactor_events
from ._base_events import *


__all__ = 'AsyncSlotBaseProactorEventLoop',


class AsyncSlotBaseProactorEventLoop(
    AsyncSlotBaseEventLoop,
    asyncio.proactor_events.BaseProactorEventLoop
):
    pass



import asyncslot
import signal
import sys
import threading
import time
import unittest
from shim import QtCore


def _raise_ki():
    return signal.default_int_handler(signal.SIGINT, None)


class TestCtrlC(unittest.TestCase):
    # A collection of test cases for the behavior of Ctrl+C for a loop
    # operating in host mode.  The expected behavior is to propagate
    # KeyboardInterrupt to the caller of run_forever.
    #
    # Adapted from test.test_asyncio.test_windows_events.ProactorLoopCtrlC
    # but also tests unix event loops.

    def setUp(self) -> None:
        if QtCore.QCoreApplication.instance() is not None:
            self.app = QtCore.QCoreApplication.instance()
        else:
            self.app = QtCore.QCoreApplication([])

    def tearDown(self) -> None:
        self.app = None

    def _test_ctrl_c(self, loop):
        def SIGINT_after_delay():
            time.sleep(0.1)
            if sys.version_info < (3, 8):
                import os
                os.kill(os.getpid(), signal.SIGINT)
            else:
                signal.raise_signal(signal.SIGINT)

        thread = threading.Thread(target=SIGINT_after_delay)
        try:
            with self.assertRaises(KeyboardInterrupt):
                loop.call_soon(thread.start)
                loop.run_forever()
        finally:
            loop.close()


@unittest.skipIf(sys.platform == 'win32', 'unix only')
class TestUnixCtrlC(TestCtrlC):
    """Test Ctrl+C under unix."""

    def test_unix_loop(self):
        self._test_ctrl_c(asyncslot.AsyncSlotDefaultEventLoop())

    def test_unix_loop_with_SIGCHLD_1(self):
        loop = asyncslot.AsyncSlotDefaultEventLoop()
        loop.add_signal_handler(signal.SIGCHLD, _raise_ki)
        self._test_ctrl_c(loop)

    def test_unix_loop_with_SIGCHLD_2(self):
        loop = asyncslot.AsyncSlotDefaultEventLoop()
        loop.add_signal_handler(signal.SIGCHLD, _raise_ki)
        loop.remove_signal_handler(signal.SIGCHLD)
        self._test_ctrl_c(loop)

    def test_unix_loop_with_SIGINT_1(self):
        loop = asyncslot.AsyncSlotDefaultEventLoop()
        loop.add_signal_handler(signal.SIGINT, _raise_ki)
        self._test_ctrl_c(loop)

    def test_unix_loop_with_SIGINT_2(self):
        loop = asyncslot.AsyncSlotDefaultEventLoop()
        loop.add_signal_handler(signal.SIGINT, _raise_ki)
        loop.remove_signal_handler(signal.SIGINT)
        self._test_ctrl_c(loop)


# The Windows Ctrl+C test is not run for Python 3.7, for two reasons:
# - First, the proactor event loop in Python 3.7 does not support being
#   interrupted by Ctrl+C; see https://github.com/python/cpython/issues/67246
# - Second, Python 3.7 does not have the raise_signal function, and
#   os.kill(SIGINT) does not work under Windows; see
#   https://stackoverflow.com/questions/35772001/how-to-handle-a-signal-sigint-on-a-windows-os-machine
@unittest.skipUnless(sys.platform == 'win32', 'windows only')
@unittest.skipUnless(sys.version_info >= (3, 8), 'requires python >= 3.8')
class TestWindowsCtrlC(TestCtrlC):
    """Test Ctrl+C under windows."""

    def test_windows_proactor_loop(self):
        self._test_ctrl_c(asyncslot.AsyncSlotProactorEventLoop())

    def test_windows_selector_loop(self):
        self._test_ctrl_c(asyncslot.AsyncSlotSelectorEventLoop())


if __name__ == '__main__':
    unittest.main()



""" gui_test_clicked.py - test asyncSlot with QAbstractButton """

import asyncio
from shim import QtCore, QtWidgets
from asyncslot import asyncSlot, AsyncSlotRunner


async def quit_later():
    await asyncio.sleep(0)
    QtWidgets.QApplication.quit()


if __name__ == '__main__':
    app = QtWidgets.QApplication([])

    button = QtWidgets.QPushButton()
    button.setText('Quit')
    button.clicked.connect(asyncSlot(quit_later))
    button.show()

    timer = QtCore.QTimer()
    timer.timeout.connect(button.click)
    timer.start()

    with AsyncSlotRunner():
        if hasattr(app, 'exec'):
            app.exec()
        else:
            app.exec_()



import asyncio
from PySide6 import QtCore
from asyncslot import AsyncSlotDefaultEventLoopPolicy


async def test():
    for name in ('Alice', 'Bob'):
        print(f'Hi, {name}!')
        await asyncio.sleep(1)
    print('Bye-bye!')


def main():
    app = QtCore.QCoreApplication()
    asyncio.set_event_loop_policy(AsyncSlotDefaultEventLoopPolicy())
    asyncio.run(test())


if __name__ == '__main__':
    main()



""" _slot.py - definition of AsyncSlot decorator """

import asyncio
import functools
import inspect
from typing import Callable, Coroutine, Set
from ._base_events import AsyncSlotBaseEventLoop


__all__ = 'asyncSlot',


# Global variable to store strong reference to AsyncSlot tasks so that they
# don't get garbage collected during execution.
_running_tasks: Set[asyncio.Task] = set()


CoroutineFunction = Callable[..., Coroutine]


def asyncSlot(fn: CoroutineFunction):  # noqa
    """ Wrap a coroutine function to make it usable as a Qt slot. """

    # TODO: support decoration on @classmethod or @staticmethod by returning
    # TODO: a wrapper method descriptor.
    if not inspect.iscoroutinefunction(fn):
        raise TypeError(f'asyncSlot cannot be applied to {fn!r} because '
                        f'it is not a coroutine function')

    # Because the wrapper's signature is (*args), PySide/PyQt will always
    # call the wrapper with the signal's (full) parameter list instead of
    # the slot's parameter list if it is shorter.  We work around this by
    # "truncating" the input parameter list if needed.
    sig = inspect.signature(fn)
    params = sig.parameters

    # Parameters come in the following order of kinds (each kind is optional):
    # [POSITIONAL_ONLY]
    # [POSITION_OR_KEYWORD]
    # [VAR_POSITIONAL]
    # [KEYWORD_ONLY]
    # [VAR_KEYWORD]
    param_count = 0
    for p in params.values():
        if p.kind == p.POSITIONAL_ONLY:
            param_count += 1
        elif p.kind == p.POSITIONAL_OR_KEYWORD:
            param_count += 1
        elif p.kind == p.VAR_POSITIONAL:
            param_count = -1
        elif p.kind == p.KEYWORD_ONLY:
            if p.default is not p.empty:
                raise TypeError(f"asyncSlot cannot be applied to {fn!r} "
                                f"because it contains keyword-only argument "
                                f"'{p.name} without default")
        elif p.kind == p.VAR_KEYWORD:
            pass  # **kwargs will always be empty
        else:
            assert False, f"unexpected parameter kind '{p.kind}'"

    @functools.wraps(fn)
    def asyncSlotWrapper(*args):
        loop = asyncio.events._get_running_loop()
        if loop is None:
            raise RuntimeError('cannot call asyncSlot without a running loop')

        if not isinstance(loop, AsyncSlotBaseEventLoop):
            raise RuntimeError(f"asyncSlot is not compatible with the "
                               f"running event loop '{loop!r}'")

        # Truncate arguments if slot expects fewer than signal provides
        if 0 <= param_count < len(args):
            coro = fn(*args[:param_count])
        else:
            coro = fn(*args)

        task = loop.run_task(coro)  # TODO: set name
        _running_tasks.add(task)
        task.add_done_callback(_running_tasks.discard)

    # fn may have been decorated with Slot() or pyqtSlot().  "Carry over"
    # the decoration if so.
    if hasattr(fn, '_slots'):  # PySide2, PySide6
        asyncSlotWrapper._slots = fn._slots  # noqa
    if hasattr(fn, '__pyqtSignature__'):  # PyQt5, PyQt6
        asyncSlotWrapper.__pyqtSignature__ = fn.__pyqtSignature__

    return asyncSlotWrapper



""" shim.py - utility functions for testing """

import os


__all__ = 'QtCore',


qt_module_name = os.getenv("TEST_QT_MODULE", "")
if qt_module_name == "":
    raise ImportError("environment variable TEST_QT_MODULE must be set")


if qt_module_name == "PyQt5":
    from PyQt5 import QtCore, QtWidgets
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot

elif qt_module_name == "PyQt6":
    from PyQt6 import QtCore, QtWidgets
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot

elif qt_module_name == "PySide2":
    from PySide2 import QtCore, QtWidgets

elif qt_module_name == "PySide6":
    from PySide6 import QtCore, QtWidgets

else:
    raise ImportError(f"unsupported TEST_QT_MODULE value: '{qt_module_name}'")



"""Define the communication interface between event loop and selector."""

from typing import Any, Optional


__all__ = '_AsyncSlotNotifier', '_AsyncSlotSelectable',


class _AsyncSlotNotifier:
    """An object responsible for the communication between an asyncslot
    event loop and a selectable object."""

    def no_result(self) -> Any:
        """Called by the selectable object if no result is immediately
        available for a select() call.  Its return value is returned
        to the caller."""
        raise NotImplementedError

    def notify(self) -> None:
        """Called by the selectable object (in a separate thread) to
        notify that result is available from the last select() call."""
        raise NotImplementedError

    def wakeup(self) -> None:
        """Called by the selectable object wake up an in-progress select()
        in the worker thread."""
        raise NotImplementedError

    def close(self) -> None:
        """Called by the event loop to close the notifier object.  After
        this call, no more notifications will be received."""
        raise NotImplementedError


class _AsyncSlotSelectable:
    """Protocol for a 'selector' that supports non-blocking select and
    notification.

    A selector may be in one of the following states:
      - IDLE   : the selector is not in BUSY or CLOSED state
      - BUSY   : the last call to select() raised AsyncSlotYield, and
                 a thread worker is waiting for IO or timeout
      - CLOSED : close() has been called

    State machine:
      - [start] --- __init__ --> IDLE
      - IDLE --- close() --> CLOSED
        IDLE --- select
                 - (IO ready, timeout == 0, or notifier is None) --> IDLE
                 - (IO not ready, timeout != 0, and notifier not None) --> BUSY
        IDLE --- set_notifier --> IDLE
      - BUSY --- (IO ready or timeout reached) --> IDLE
        BUSY --- set_notifier --> (wakes up selector) --> IDLE
      - CLOSED --- [end]
    """

    def set_notifier(self, notifier: Optional[_AsyncSlotNotifier]) -> None:
        """Set the notifier.

        If the selector is in BUSY state, wake it up and wait for it
        to become IDLE before returning.  In this case, the previous
        installed notifier (if any) is still signaled.
        """
        raise NotImplementedError

    def select(self, timeout: Optional[float] = None):
        """
        If timeout is zero or some IO is readily available, return the
        available IO immediately.

        If timeout is not zero, IO is not ready and notifier is not None,
        launch a thread worker to perform the real select() and return
        notifier.no_result(), which possibly throws.  When the real select()
        completes, signal the notifier object.

        If timeout is not zero, IO is not ready and notifier is None,
        perform normal (blocking) select.
        """
        raise NotImplementedError



""" sleep_example.py - simple GUI to test async slot """

import asyncio
import sys
import time
from PyQt6 import QtWidgets
from bouncingwidget import BouncingWidget
from asyncslot import asyncSlot, AsyncSlotRunner
import traceback


def wrap(fn):
    def call(*args, **kwargs):
        try:
            fn(*args, **kwargs)
        except BaseException as exc:
            print(traceback.format_exc(), file=sys.stderr)
            raise
    return call


class MyWidget(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("AsyncSlot Test")

        self._layout = QtWidgets.QVBoxLayout(self)
        self._bouncer = BouncingWidget()
        self._layout.addWidget(self._bouncer)
        self._button_group = QtWidgets.QHBoxLayout()
        self._sync_button = QtWidgets.QPushButton("Sync Sleep")
        self._async_button = QtWidgets.QPushButton("Async Sleep")
        self._button_group.addWidget(self._sync_button)
        self._button_group.addWidget(self._async_button)
        self._layout.addLayout(self._button_group)

        self._sync_button.clicked.connect(self.sync_sleep)
        self._async_button.clicked.connect(asyncSlot(self.async_sleep))

    def sync_sleep(self):
        time.sleep(3)
        QtWidgets.QMessageBox.information(self, 'Test', 'Done')

    async def async_sleep(self, checked):
        self._async_button.setEnabled(False)
        await asyncio.sleep(3)
        QtWidgets.QMessageBox.information(self, 'Test', 'Done')
        self._async_button.setEnabled(True)


def main():
    app = QtWidgets.QApplication([])

    widget = MyWidget()
    widget.resize(400, 300)
    widget.show()

    with AsyncSlotRunner():
        sys.exit(app.exec())


if __name__ == "__main__":
    main()



from shim import QtCore
import asyncio
import inspect
import time
import unittest
import asyncslot


def print_stack():
    for i, o in enumerate(inspect.stack(0)):
        if i > 0:
            print(o)


def get_call_stack():
    return [o.function for o in inspect.stack(0)[1:]]


class TestRunTask(unittest.TestCase):
    # run_task should execute the task immediately until the first yield.

    def setUp(self) -> None:
        if QtCore.QCoreApplication.instance() is not None:
            self.app = QtCore.QCoreApplication.instance()
        else:
            self.app = QtCore.QCoreApplication([])
        self.loop = asyncslot.AsyncSlotDefaultEventLoop()

    def tearDown(self) -> None:
        self.loop.close()
        self.app = None

    def test_no_yield_benchmark(self):
        # create_task with coroutine with no yield is not executed eagerly
        async def coro(output):
            output.append(1)
            return 'finished'

        async def as_test_entry():
            output = []
            task = self.loop.create_task(coro(output))
            self.assertFalse(task.done())
            self.assertEqual(output, [])
            value = await task
            self.assertEqual(output, [1])
            return value

        result = self.loop.run_until_complete(as_test_entry())
        self.assertEqual(result, 'finished')

    def test_no_yield(self):
        # coroutine with no yield should be eagerly executed to completion
        async def coro():
            return get_call_stack()

        async def as_test_entry():
            task = self.loop.run_task(coro())
            self.assertTrue(task.done())
            return task.result()

        result = self.loop.run_until_complete(as_test_entry())
        self.assertIn('as_test_entry', result)

    def test_one_yield(self):
        # coroutine with one yield should be eagerly executed
        async def coro(output):
            output.append(1)
            await asyncio.sleep(0)
            return get_call_stack()

        async def as_test_entry():
            output = []
            task = self.loop.run_task(coro(output))
            self.assertEqual(output, [1])
            self.assertFalse(task.done())
            return await task

        result = self.loop.run_until_complete(as_test_entry())
        self.assertNotIn('as_test_entry', result)

    def test_interleaved(self):
        # run_task interleaved with create_task should work correctly
        var = 10

        async def coro1():
            nonlocal var
            var += 6
            await asyncio.sleep(0)
            var += 7

        async def coro2():
            nonlocal var
            var /= 8

        async def as_test_entry():
            task2 = self.loop.create_task(coro2())
            task1 = self.loop.run_task(coro1())
            await asyncio.gather(task1, task2)

        self.loop.run_until_complete(as_test_entry())
        self.assertEqual(var, 9)

    def test_current_task_before_yield(self):
        pass

    def test_current_task_after_yield(self):
        pass

    def test_current_task_before_yield_no_loop(self):
        pass

    def test_run_task_3(self):
        # running an async generator should raise an error (or not?)
        pass

    def test_raise_before_yield(self):
        # An exception raised before yield should be propagated to the caller
        pass

    def test_raise_after_yield(self):
        # Exceptions raised after yield should be treated normally
        pass

    def test_cancellation(self):
        # The returned task may be cancelled
        pass

    def test_nested(self):
        # run_task may run_task again, and still be immediate execution
        pass

    def test_recursive_ok(self):
        # run_task recursively
        pass

    def test_recursive_error(self):
        # run_task recursively too deeply should raise StackOverflowError
        pass


class TestRunTaskWithoutRunningLoop(unittest.TestCase):

    def setUp(self) -> None:
        if QtCore.QCoreApplication.instance() is None:
            self.app = QtCore.QCoreApplication([])
        self.loop = asyncslot.AsyncSlotDefaultEventLoop()

    def tearDown(self) -> None:
        self.loop.close()

    def test_no_yield_no_loop(self):
        # run_task with no yield and no running loop should be ok
        async def coro():
            return 'magic'

        task = self.loop.run_task(coro())
        self.assertTrue(task.done())
        self.assertEqual(task.result(), 'magic')

    def test_one_yield_no_loop(self):
        # run_task with no running loop should still execute the first step.
        state = 'initial'

        async def coro():
            nonlocal state
            state = 'executed'
            await asyncio.sleep(0)
            state = 'finished'

        task = self.loop.run_task(coro())
        self.assertEqual(state, 'executed')
        self.assertFalse(task.done())
        # Finish the suspended task
        self.loop.run_until_complete(task)
        self.assertTrue(task.done())
        self.assertEqual(state, 'finished')


if __name__ == '__main__':
    unittest.main()



""" minimal_gui.py - minimal gui application using asyncslot """

import asyncio
from PySide6 import QtWidgets
from asyncslot import asyncSlot, AsyncSlotRunner


async def say_hi():
    await asyncio.sleep(1)
    QtWidgets.QMessageBox.information(None, "Demo", "Hi")


if __name__ == '__main__':
    app = QtWidgets.QApplication()

    button = QtWidgets.QPushButton()
    button.setText('Say Hi after one second')
    button.clicked.connect(asyncSlot(say_hi))
    button.show()

    with AsyncSlotRunner():
        app.exec()



""" bindings.py - resolve Python/Qt binding at run-time """

import os
import sys


__all__ = 'QtCore',


imported = []
for binding in ('PyQt5', 'PyQt6', 'PySide2', 'PySide6'):
    if binding in sys.modules:
        imported.append(binding)


if len(imported) == 0:
    binding = os.getenv("ASYNCSLOT_QTBINDING", "")
    if not binding:
        raise ImportError(
            'no Qt binding is imported and ASYNCSLOT_QTBINDING is not set')

elif len(imported) == 1:
    binding = imported[0]

else:
    raise ImportError(f'more than one Qt bindings are imported: {imported}')


# Explicitly list the branches for coverage testing.
if binding == 'PyQt5':
    from PyQt5 import QtCore

elif binding == 'PyQt6':
    from PyQt6 import QtCore

elif binding == 'PySide2':
    from PySide2 import QtCore

elif binding == 'PySide6':
    from PySide6 import QtCore

else:
    raise ImportError(f"unsupported ASYNCSLOT_QTBINDING value '{binding}'")



"""asyncio_tests.py - run Python bundled asyncio test suite

The test suite is located in Lib/test/test_asyncio.  Available test modules
under different Python versions are as follows:

Module                      3.7    3.8    3.9    3.10    3.11
-------------------------------------------------------------
test_base_events.py         ✓      ✓      ✓      ✓       ✓
test_buffered_proto.py      ✓      ✓      ✓      ✓       ✓
test_context.py             ✓      ✓      ✓      ✓       ✓
test_events.py              ✓      ✓      ✓      ✓       ✓
test_futures.py             ✓      ✓      ✓      ✓       ✓
test_locks.py               ✓      ✓      ✓      ✓       ✓
test_pep492.py              ✓      ✓      ✓      ✓       ✓
test_proactor_events.py     ✓      ✓      ✓      ✓       ✓
test_queues.py              ✓      ✓      ✓      ✓       ✓
test_runners.py             ✓      ✓      ✓      ✓       ✓
test_selector_events.py     ✓      ✓      ✓      ✓       ✓
test_server.py              ✓      ✓      ✓      ✓       ✓
test_sslproto.py            ✓      ✓      ✓      ✓       ✓
test_streams.py             ✓      ✓      ✓      ✓       ✓
test_subprocess.py          ✓      ✓      ✓      ✓       ✓
test_tasks.py               ✓      ✓      ✓      ✓       ✓
test_transports.py          ✓      ✓      ✓      ✓       ✓
test_unix_events.py         ✓      ✓      ✓      ✓       ✓
test_windows_events.py      ✓      ✓      ✓      ✓       ✓
test_windows_utils.py       ✓      ✓      ✓      ✓       ✓
test_asyncio_waitfor.py            ✓
test_futures2.py                   ✓      ✓      ✓       ✓
test_protocols.py                  ✓      ✓      ✓       ✓
test_sendfile.py                   ✓      ✓      ✓       ✓
test_sock_lowlevel.py              ✓      ✓      ✓       ✓
test_threads.py                           ✓      ✓       ✓
test_waitfor.py                           ✓      ✓       ✓
test_ssl.py                                              ✓
test_taskgroups.py                                       ✓
test_timeouts.py                                         ✓

"""
import sys

import asyncio
import asyncio.base_events
import asyncio.selector_events
import asyncio.proactor_events
if sys.platform == 'win32':
    import asyncio.windows_events
else:
    import asyncio.unix_events

import asyncslot
import unittest

from asyncslot.bindings import QtCore
app = QtCore.QCoreApplication([])

# We now need to monkey-patch asyncio ...

asyncio.BaseEventLoop = asyncio.base_events.BaseEventLoop = asyncslot.AsyncSlotBaseEventLoop
asyncio.selector_events.BaseSelectorEventLoop = asyncslot.AsyncSlotBaseSelectorEventLoop
asyncio.proactor_events.BaseProactorEventLoop = asyncslot.AsyncSlotBaseProactorEventLoop

if sys.platform == 'win32':
    asyncio.SelectorEventLoop = asyncio.windows_events.SelectorEventLoop = asyncio.windows_events._WindowsSelectorEventLoop = asyncslot.AsyncSlotSelectorEventLoop
    asyncio.ProactorEventLoop = asyncio.windows_events.ProactorEventLoop = asyncslot.AsyncSlotProactorEventLoop
    asyncio.IocpProactor = asyncio.windows_events.IocpProactor = asyncslot._windows_events._AsyncSlotProactor
    asyncio.WindowsSelectorEventLoopPolicy = asyncio.windows_events.WindowsSelectorEventLoopPolicy = asyncslot.AsyncSlotSelectorEventLoopPolicy
    asyncio.WindowsProactorEventLoopPolicy = asyncio.windows_events.WindowsProactorEventLoopPolicy = asyncslot.AsyncSlotProactorEventLoopPolicy
    asyncio.DefaultEventLoopPolicy = asyncio.windows_events.DefaultEventLoopPolicy = asyncslot.AsyncSlotDefaultEventLoopPolicy
else:
    asyncio.SelectorEventLoop = asyncio.unix_events.SelectorEventLoop = asyncio.unix_events._UnixSelectorEventLoop = asyncslot.AsyncSlotSelectorEventLoop
    asyncio.DefaultEventLoopPolicy = asyncio.unix_events.DefaultEventLoopPolicy = asyncio.unix_events._UnixDefaultEventLoopPolicy = asyncslot.AsyncSlotDefaultEventLoopPolicy


# Now import the tests into __main__
from test.test_asyncio import load_tests

# The following test is expected to fail because the call stack is
# changed under asyncslot.  If the test succeeds, it means the monkey
# patching didn't work!
import test.test_asyncio.test_events
test.test_asyncio.test_events.HandleTests.test_handle_source_traceback = \
    unittest.expectedFailure(test.test_asyncio.test_events.HandleTests.test_handle_source_traceback)

# To run a particular test, import that test class, and specify
# TestClassName.test_name on the command line. For example:
# python asyncio_tests.py ProactorLoopCtrlC.test_ctrl_c

# TODO: why do we display warnings to stderr, but not asyncio?

if __name__ == "__main__":
    unittest.main()



""" dep_test_del.py - demo a strange bug with PyQt6, PySide2 and PySide6

With certain code (unrelated to asyncslot), these bindings raise an
exception in asyncio's event loop's __del__ method, complaining about
invalid file handle when attempting to unregister a self-read socket.
It appears that a reference cycle is created if a QtCore.QCoreApplication
instance is created.
"""

import asyncio
import concurrent.futures
import importlib
import selectors
import os

qt_binding_name = os.getenv("TEST_QT_BINDING", "")
if not qt_binding_name:
    raise RuntimeError("TEST_QT_BINDING must be specified")

QtCore = importlib.import_module(f"{qt_binding_name}.QtCore")


async def noop():
    pass


class CustomSelector:
    def __init__(self):
        self._selector = selectors.DefaultSelector()
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        executor.shutdown(wait=True)
        executor = None

    def register(self, fileobj, events, data=None):
        return self._selector.register(fileobj, events, data)

    def unregister(self, fileobj):
        return self._selector.unregister(fileobj)

    def modify(self, fileobj, events, data=None):
        return self._selector.modify(fileobj, events, data)

    def select(self, timeout):
        return self._selector.select(timeout)

    def close(self):
        self._selector.close()

    def get_key(self, fileobj):
        return self._selector.get_key(fileobj)

    def get_map(self):
        return self._selector.get_map()


class CustomEventLoop(asyncio.SelectorEventLoop):
    def __init__(self):
        whatever = CustomSelector()
        whatever = None
        selector = selectors.DefaultSelector()
        super().__init__(selector)


def main():
    app = QtCore.QCoreApplication([])
    loop = CustomEventLoop()
    loop.run_until_complete(noop())


if __name__ == '__main__':
    main()



from shim import QtCore
import asyncio
import time
import unittest
import asyncslot


class TestSleep(unittest.TestCase):

    def setUp(self) -> None:
        if QtCore.QCoreApplication.instance() is not None:
            self.app = QtCore.QCoreApplication.instance()
        else:
            self.app = QtCore.QCoreApplication([])
        self.loop = asyncslot.AsyncSlotDefaultEventLoop()

    def tearDown(self) -> None:
        self.loop.close()
        self.app = None

    def test_sleep(self):
        total_duration = 2.0

        async def sleep_for(times):
            for _ in range(times):
                await asyncio.sleep(total_duration / times)

        async def entry():
            t1 = time.time()
            tasks = []
            for _ in range(100):
                tasks.append(sleep_for(8))
                tasks.append(sleep_for(4))
                tasks.append(sleep_for(2))
                tasks.append(sleep_for(1))
            await asyncio.gather(*tasks)
            t2 = time.time()
            return t2 - t1

        total_time = self.loop.run_until_complete(entry())
        self.assertGreater(total_time, total_duration - 0.1)
        self.assertLess(total_time, total_duration + 1.0)


if __name__ == '__main__':
    unittest.main()



""" _selector_events.py - AsyncSlot based on SelectorEventLoop """

import asyncio.selector_events
import concurrent.futures
import selectors
import signal
import threading
import unittest.mock
from typing import List, Optional, Tuple
from ._base_events import *
from ._selectable import _AsyncSlotNotifier


__all__ = 'AsyncSlotBaseSelectorEventLoop',


class _AsyncSlotSelector(selectors.BaseSelector):

    def __init__(self, selector: selectors.BaseSelector):
        super().__init__()
        self._selector = selector
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self._select_future: Optional[concurrent.futures.Future] = None
        self._idle = threading.Event()
        self._idle.set()
        self._notifier: Optional[_AsyncSlotNotifier] = None
        self._closed = False

    def set_notifier(self, notifier: Optional[_AsyncSlotNotifier]) -> None:
        self._unblock_if_blocked()
        self._notifier = notifier

    def _unblock_if_blocked(self):
        assert not self._closed, 'selector already closed'
        if not self._idle.is_set():
            assert self._notifier is not None, 'notifier expected'
            self._notifier.wakeup()
            self._idle.wait()

    def register(self, fileobj, events, data=None):
        self._unblock_if_blocked()
        return self._selector.register(fileobj, events, data)

    def unregister(self, fileobj):
        self._unblock_if_blocked()
        return self._selector.unregister(fileobj)

    def modify(self, fileobj, events, data=None):
        self._unblock_if_blocked()
        return self._selector.modify(fileobj, events, data)

    def select(self, timeout: Optional[float] = None) \
            -> List[Tuple[selectors.SelectorKey, int]]:
        assert not self._closed, 'selector already closed'

        # If the last call to select() raised AsyncSlotYield, the caller
        # (from _run_once) should only call us again after receiving a
        # notification from us, and we only send the notification after
        # entering IDLE state.
        assert self._idle.is_set(), 'unexpected select'

        # Return previous select() result (or exception) if there is one.
        if self._select_future is not None:
            try:
                return self._select_future.result()
            finally:
                self._select_future = None

        # Perform normal (blocking) select if no notifier is set.
        if self._notifier is None:
            return self._selector.select(timeout)

        # Perform normal select if timeout is zero.
        if timeout == 0:
            return self._selector.select(timeout)

        # Try select with zero timeout, and return if any IO is ready.
        event_list = self._selector.select(0)
        if event_list:
            return event_list

        # No IO is ready and caller wants to wait.  select() in a separate
        # thread and tell the caller to yield.
        self._idle.clear()
        try:
            self._select_future = self._executor.submit(self._select, timeout)
        except BaseException:
            # Should submit() raise, we assume no task is spawned.
            self._idle.set()
            raise
        else:
            return self._notifier.no_result()  # raises _AsyncSlotYield

    def _select(self, timeout):
        try:
            return self._selector.select(timeout)
        finally:
            # Make a copy of self._notifier because it may be altered by
            # set_notifier immediately after self._idle is set.
            notifier = self._notifier
            self._idle.set()
            notifier.notify()

    def close(self) -> None:
        # close() is called when the loop is being closed, and the loop
        # can only be closed when it is in STOPPED state.  In this state
        # the selector must be idle.  In addition, the self pipe is
        # closed before closing the selector, so write_to_self cannot be
        # used at this point.
        if not self._closed:
            assert self._idle.is_set(), 'unexpected close'
            self._executor.shutdown()
            self._selector.close()
            self._select_future = None
            self._notifier = None
            self._closed = True

    def get_key(self, fileobj):
        self._unblock_if_blocked()
        return self._selector.get_key(fileobj)

    def get_map(self):
        self._unblock_if_blocked()
        return self._selector.get_map()


class AsyncSlotBaseSelectorEventLoop(
    AsyncSlotBaseEventLoop,
    asyncio.selector_events.BaseSelectorEventLoop
):
    def __init__(self, selector=None):
        if selector is None:
            selector = selectors.DefaultSelector()
        if isinstance(selector, unittest.mock.Mock):
            # Pass through mock object for testing
            asyncslot_selector = selector
        else:
            asyncslot_selector = _AsyncSlotSelector(selector)
        super().__init__(asyncslot_selector)

        # Similar to asyncio.BaseProactorEventLoop, install wakeup fd
        # so that select() in a separate thread can be interrupted by
        # Ctrl+C.  Only the main thread of the main interpreter may
        # install a wakeup fd, but other threads will never receive a
        # KeyboardInterrupt, so it's ok if set_wakeup_fd fails.
        self.__wakeup_fd_installed = False
        self._asyncslot_install_wakeup_fd()

    def _asyncslot_install_wakeup_fd(self):
        try:
            signal.set_wakeup_fd(self._csock.fileno())
        except Exception:
            self.__wakeup_fd_installed = False
        else:
            self.__wakeup_fd_installed = True

    def close(self):
        if self.is_running():
            raise RuntimeError("Cannot close a running event loop")
        # Uninstall the wakeup fd is one was installed.
        if self.__wakeup_fd_installed:
            try:
                signal.set_wakeup_fd(-1)
            except (ValueError, OSError):
                pass
            finally:
                self.__wakeup_fd_installed = False
        super().close()



""" _runners.py - run asyncslot event loop in INTEGRATED mode """

import asyncio.runners
import sys
from typing import Callable, Optional
from ._base_events import AsyncSlotBaseEventLoop

if sys.platform == 'win32':
    from ._windows_events import AsyncSlotDefaultEventLoop
else:
    from ._unix_events import AsyncSlotDefaultEventLoop


__all__ = 'AsyncSlotRunner',


# Adapted from asyncio.runners
class AsyncSlotRunner:
    """Context manager that runs an AsyncSlotDefaultEventLoop in INTEGRATE
    mode."""

    def __init__(self, *, debug: Optional[bool] = None, loop_factory:
                 Optional[Callable[[], AsyncSlotBaseEventLoop]] = None):
        self._debug = debug
        self._loop_factory = loop_factory
        self._loop: Optional[AsyncSlotBaseEventLoop] = None

    def __enter__(self):
        if self._loop_factory is None:
            self._loop = AsyncSlotDefaultEventLoop()
        else:
            self._loop = self._loop_factory()
        self._loop.set_guest(True)
        asyncio.events.set_event_loop(self._loop)  # ???
        if self._debug is not None:
            self._loop.set_debug(self._debug)
        self._loop.start()

    def __exit__(self, *args):
        if self._loop.is_running():
            # Don't stop again if user code has already stopped the loop.
            self._loop.stop()
        loop = self._loop
        try:
            asyncio.runners._cancel_all_tasks(loop)
            # Note: the following steps will be run in EXCLUSIVE mode as
            # it is undesirable, and maybe even impossible, to launch a
            # Qt event loop at this point -- e.g. QCoreApplication.exit()
            # may have been called.
            loop.run_until_complete(loop.shutdown_asyncgens())
            if hasattr(loop, "shutdown_default_executor"):
                loop.run_until_complete(loop.shutdown_default_executor())
        finally:
            asyncio.events.set_event_loop(None)
            loop.close()
            self._loop = None



"""Implements asyncio event loop based on Qt event loop.

The asyncio event loop class hierarchy is as follows:

class                             submodule        alias
------------------------------------------------------------------------------
BaseEventLoop                     base_events
  BaseSelectorEventLoop           selector_events
    _UnixSelectorEventLoop        unix_events      SelectorEventLoop [1]
    _WindowsSelectorEventLoop     windows_events   SelectorEventLoop [2,3]
  BaseProactorEventLoop           proactor_events
    ProactorEventLoop             windows_events
BaseDefaultEventLoopPolicy        events
  _UnixDefaultEventLoopPolicy     unix_events      DefaultEventLoopPolicy [1]
  WindowsSelectorEventLoopPolicy  windows_events   DefaultEventLoopPolicy [2]
  WindowsProactorEventLoopPolicy  windows_events   DefaultEventLoopPolicy [3]

[1] under unix
[2] under Windows, for Python 3.7
[3] under Windows, for Python 3.8 and above

For ease of reference and to facilitate testing, asyncslot's source code is
arrange in a similar structure:

class                               submodule         alias
------------------------------------------------------------------------------
AsyncSlotBaseEventLoop              _base_events
  AsyncSlotBaseSelectorEventLoop    _selector_events
    AsyncSlotSelectorEventLoop      _unix_events      AsyncSlotDefaultEventLoop [1]
    AsyncSlotSelectorEventLoop      _windows_events   AsyncSlotDefaultEventLoop [2]
  AsyncSlotBaseProactorEventLoop    _proactor_events
    AsyncSlotProactorEventLoop      _windows_events   AsyncSlotDefaultEventLoop [3]
(asyncio.events.BaseDefaultEventLoopPolicy)
  AsyncSlotSelectorEventLoopPolicy  _unix_events      AsyncSlotDefaultEventLoopPolicy [1]
  AsyncSlotSelectorEventLoopPolicy  _windows_events   AsyncSlotDefaultEventLoopPolicy [2]
  AsyncSlotProactorEventLoopPolicy  _windows_events   AsyncSlotDefaultEventLoopPolicy [3]

"""
import sys

from ._base_events import *
from ._selector_events import *
from ._proactor_events import *
from ._slot import *
from ._runners import *


__all__ = (
    _base_events.__all__ +
    _selector_events.__all__ +
    _proactor_events.__all__ +
    _slot.__all__ +
    _runners.__all__
)


if sys.platform == 'win32':
    from ._windows_events import *
    __all__ += _windows_events.__all__
else:
    from ._unix_events import *
    __all__ += _unix_events.__all__
